import json
import os.path

import faker

import stix2generator.language.builder
import stix2generator.generation.object_generator
import stix2generator.generation.reference_graph_generator
import stix2generator.generation.semantics
import stix2generator.generation.stix_generator
from stix2generator.exceptions import RegistryNotFoundError

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

_STIX_REGISTRIES = {
    "2.1": "stix21_registry.json"
}


def _get_registry(stix_version):
    """
    Get the object generator registry for the given STIX version.

    :param stix_version: a STIX version
    :return: An object generator registry, as parsed JSON
    :raises stix2generator.exceptions.RegistryNotFoundError: If there isn't a
        built-in registry for the given STIX version
    :raises IOError: (python 2) If the registry JSON file couldn't be opened
        or read
    :raises OSError: (python 3) If the registry JSON file couldn't be opened
        or read (IOError is retained as an alias for backward compatibility).
    """
    if stix_version not in _STIX_REGISTRIES:
        raise RegistryNotFoundError(stix_version)

    spec_registry_path = os.path.join(
        os.path.dirname(__file__),
        _STIX_REGISTRIES[stix_version]
    )

    with open(spec_registry_path, encoding="utf-8") as f:
        spec_registry = json.load(f)

    return spec_registry


def create_object_generator(
    object_generator_config=None, extra_specs=None, stix_version="2.1"
):
    """
    Create an instance of the built-in object generator.

    :param object_generator_config: A Config object with generator settings,
        or None to use defaults.
    :param extra_specs: Any extra specifications which should be added to the
        built-in specs this generator will use, or None if there aren't any.
        The extra specs are overlaid on the built-in specs, so built-in specs
        can be overridden this way. (dict)
    :param stix_version: The STIX version to use, as a string ("2.0", "2.1",
        etc).
    :return: The object generator
    :raises stix2generator.exceptions.RegistryNotFoundError: If there isn't a
        built-in registry for the given STIX version
    :raises IOError: (python 2) If the registry JSON file couldn't be opened
        or read
    :raises OSError: (python 3) If the registry JSON file couldn't be opened
        or read (IOError is retained as an alias for backward compatibility).
    """

    spec_registry = _get_registry(stix_version)

    if extra_specs:
        _update_dict_recursive(spec_registry, extra_specs)

    stix_semantics = stix2generator.generation.semantics.STIXSemantics()

    faker_ = faker.Faker()
    faker_semantics = stix2generator.generation.semantics.FakerSemantics(faker_)

    semantics_providers = [
        stix_semantics,
        faker_semantics
    ]

    generator = stix2generator.generation.object_generator.ObjectGenerator(
        spec_registry,
        semantics_providers,
        object_generator_config
    )

    return generator


def create_stix_generator(
    object_generator_config=None,
    ref_graph_generator_config=None,
    stix_generator_config=None,
    extra_specs=None,
    stix_version="2.1"
):
    object_generator = create_object_generator(
        object_generator_config, extra_specs, stix_version
    )

    ref_graph_generator = stix2generator.generation.reference_graph_generator \
        .ReferenceGraphGenerator(
            object_generator, ref_graph_generator_config, stix_version
        )

    stix_generator = stix2generator.generation.stix_generator.STIXGenerator(
        object_generator, ref_graph_generator, stix_generator_config,
        stix_version
    )

    return stix_generator


def _update_dict_recursive(base_dict, new_dict):
    """
    Recursively updates a dictionary.

    :param base_dict: The original dictionary that will be updated in place.
    :param new_dict: The dictionary of values to update the original one with.
    :return: The updated dictionary
    """

    for key, val in new_dict.items():
        if isinstance(val, Mapping):
            base_dict[key] = _update_dict_recursive(base_dict.get(key, {}), val)
        else:
            base_dict[key] = val
    return base_dict


def create_default_language_processor(
    object_generator_config=None,
    extra_specs=None,
    stix_version="2.1"
):
    """
    Creates a language processor instance which uses the built-in object
    generator with the given settings.

    :param object_generator_config: Configuration information for the
        object generator, which will be used to generate any objects
        required by prototyping language this processor processes.
    :param extra_specs: Any extra specifications which should be added to
        the built-in specs this generator will use, or None if there aren't
        any.  The extra specs are overlaid on the built-in specs, so
        built-in specs can be overridden this way. (dict)
    :param stix_version: Which version of STIX to use.
    :return: A language processor instance
    """
    object_generator = create_object_generator(
        object_generator_config,
        extra_specs,
        stix_version
    )

    processor = stix2generator.language.builder.LanguageProcessor(
        object_generator, stix_version
    )

    return processor



import collections
import collections.abc
import itertools
import random
import stix2.utils

import stix2generator.exceptions
import stix2generator.generation
import stix2generator.generation.reference_graph_generator
import stix2generator.generation.relationships
import stix2generator.utils


class Config(stix2generator.generation.Config):
    """
    STIX generator config settings.

    min/max_relationships: Limits on the number of SROs created.  This
        influences (but does not solely determine) the size of the resulting
        graph.

    probability_reuse: One endpoint of every SRO connects to existing content,
        to avoid partitioning the graph.  This setting determines how likely
        the other endpoint(s) are to refer to existing content.

    probability_sighting: When each SRO is added, with this probability, a
        sighting will be attempted instead of a "plain" relationship.  A plain
        relationship is still used as a fallback if a sighting cannot be
        created.

    complete_ref_properties: When an object is created, whether or not to
        leave its reference properties dangling.  Applies only to
        non-SRO-endpoint reference properties.
    """
    _DEFAULTS = {
        "min_relationships": 3,
        "max_relationships": 6,
        "probability_reuse": 0.5,
        "probability_sighting": 0.2,
        "complete_ref_properties": True
    }


def _random_id_of_types(by_type, *types, stix_version="2.1"):
    """
    Choose a random ID of one of the given types, from the by_type mapping.

    :param by_type: A mapping from STIX type to a list of STIX IDs of that type
    :param types: The STIX types of interest.  If no types are given, it means
        no types are legal, so None will always be returned.
    :param stix_version: The STIX version used to evaluate constraints
    :return: A STIX ID of one of the given types, or None if by_types is empty
        or does not contain any IDs of any of the required types.
    """

    # This is a bit more complicated, but produces a uniform distribution over
    # the IDs.  A simpler way would be to choose a random type first, and then
    # a random ID of the chosen type, but that would not result in a uniform
    # distribution over IDs.

    num_ids = sum(
        len(ids)
        for type_, ids in by_type.items()
        if stix2generator.utils.is_stix_type(
            type_, *types, stix_version=stix_version
        )
    )

    if num_ids == 0:
        id_ = None

    else:
        id_iter = itertools.chain.from_iterable(
            ids
            for type_, ids in by_type.items()
            if stix2generator.utils.is_stix_type(
                type_, *types, stix_version=stix_version
            )
        )

        id_ = stix2generator.utils.rand_iterable(
            id_iter, num_ids
        )

    return id_


def _is_sro_connectable(obj_or_type, stix_version):
    """
    Checks whether the given STIX type, or the type of the given STIX object
    is "connectable" via (plain, non-sighting) SRO.  I.e. can it appear as
    the source or target ref of an SRO.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param stix_version: A STIX version as a string
    :return: True if the type is connectable via SRO; False if not
    """
    return stix2generator.utils.is_stix_type(
        obj_or_type,
        stix2generator.utils.STIXTypeClass.SDO,
        stix2generator.utils.STIXTypeClass.SCO,
        stix_version=stix_version
    )


def _random_sro_connectable_id(by_type, stix_version):
    """
    Choose a random ID of an SRO-connectable type (see _is_sro_connectable())
    from the by_type map.

    :param by_type: A map from STIX type to a list of IDs of STIX objects in
        the current graph of that type
    :param stix_version: A STIX version as a string
    :return: An SRO-connectable STIX ID
    """
    return _random_id_of_types(
        by_type,
        stix2generator.utils.STIXTypeClass.SDO,
        stix2generator.utils.STIXTypeClass.SCO,
        stix_version=stix_version
    )


class STIXGenerator:
    """
    Instances generate STIX content: STIX objects related by SROs and
    embedded relationships.
    """

    def __init__(
        self, object_generator, ref_graph_generator=None, config=None,
        stix_version="2.1"
    ):
        """
        Initialize this STIX generator.

        :param object_generator: An instance of the built-in object generator
            to use for generating the objects.
        :param ref_graph_generator: A reference graph generator used if
            dangling references must be avoided.  If None, automatically create
            one with default config settings.
        :param config: A Config object with settings for this generator.
        :param stix_version: A STIX version, as a string.
        """
        self.__object_generator = object_generator
        self.__ref_graph_generator = \
            ref_graph_generator \
            or stix2generator.generation.reference_graph_generator\
            .ReferenceGraphGenerator(self.__object_generator)
        self.__config = config or Config()
        self.__stix_version = stix_version

    @property
    def config(self):
        return self.__config

    def __random_generatable_stix_type(self, *constraints):
        """
        Wraps stix2generator.utils.random_generatable_stix_type() and does some
        error handling, to avoid repeating it at various call sites.  See the
        aforementioned function for more information.

        :param constraints: Constraints describing STIX types
        :return: A satisfying generatable STIX type
        :raises stix2generator.exceptions.GeneratableSTIXTypeNotFoundError: if
            a satisfying generatable STIX type could not be found
        """
        stix_type = stix2generator.utils.random_generatable_stix_type(
            self.__object_generator, *constraints,
            stix_version=self.__stix_version
        )

        if not stix_type:
            raise stix2generator.exceptions.GeneratableSTIXTypeNotFoundError(
                constraints, self.__stix_version
            )

        return stix_type

    def __make_sro(self, rel, id1, id2, type1=None, type2=None):
        """
        Make an SRO dict of the type reflected by the given Relationship object
        (rel), which connects the given two IDs together.  If the SRO could
        "point" in either direction (id1 -> id2 or id2 -> id1), then a
        direction is chosen at random.

        :param rel: A Relationship object, which gives the required source and
            target types, and the relationship type
        :param id1: The STIX ID of one SRO endpoint
        :param id2: The STIX ID of the other SRO endpoint
        :param type1: The type of id1 if known, None if unknown.  If not
            passed, it is derived from id1.  Prevents having to recompute it,
            if it is already known at a call site.
        :param type2: The type of id2, analogous to type1
        :return: The SRO dict
        """

        if not type1:
            type1 = stix2.utils.get_type_from_id(id1)

        if not type2:
            type2 = stix2.utils.get_type_from_id(id1)

        if type1 == type2:
            # let it be 50-50 which way we point the SRO, if it can
            # point either way.
            source_ref = id1
            target_ref = id2
            if random.random() < 0.5:
                source_ref, target_ref = target_ref, source_ref

        elif type1 == rel.src_type:
            source_ref = id1
            target_ref = id2

        else:
            source_ref = id2
            target_ref = id1

        sro = self.__object_generator.generate("relationship")
        sro["source_ref"] = source_ref
        sro["target_ref"] = target_ref
        sro["relationship_type"] = rel.rel_type

        return sro

    def __make_related_to_sro(self, id1, id2):
        """
        Make a "related-to" SRO connecting id1 and id2.  The directionality of
        the SRO will be randomly chosen.

        :param id1: The STIX ID of an endpoint
        :param id2: The STIX ID of another endpoint
        :return: A related-to SRO dict
        """

        source_ref = id1
        target_ref = id2
        if random.random() < 0.5:
            source_ref, target_ref = target_ref, source_ref

        sro = self.__object_generator.generate("relationship")
        sro["source_ref"] = source_ref
        sro["target_ref"] = target_ref
        sro["relationship_type"] = "related-to"

        return sro

    def __make_ref_graph(self, seed_type, by_id, by_type):
        """
        Make a reference graph seeded with an object of the given type, and
        integrate it into the current STIX graph.  Return the ID of the
        generated seed object.  The graph bookkeeping structures are updated
        in-place.

        :param seed_type: A STIX type
        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        :return: The ID of the generated seed object
        """

        # This is an easy way of being able to see a "delta" between the old
        # content and new.  We need to make a pass through the new content to
        # update our bookkeeping.  Ought to be more efficient than scanning the
        # whole graph each time.
        chained_graph = collections.ChainMap(by_id).new_child()

        seed_id, _ = self.__ref_graph_generator.generate(
            seed_type, chained_graph
        )

        new_content = chained_graph.maps[0]
        by_id.update(new_content)

        for id_, obj in new_content.items():
            by_type.setdefault(obj["type"], []).append(id_)

        return seed_id

    def __complete_ref_properties(self, obj, by_id, by_type):
        """
        Find all ref properties of the given object, and generate additional
        objects and fix up the references such that none are dangling.  The
        object and bookkeeping structures are updated in-place.

        This uses the reference graph generator, which will deal with things
        like inverse property constraints, to ensure a more semantically
        sensible graph.  In most cases, to be safe, this is probably how one
        should deal with reference properties.  In specific cases, where
        there is more assurance that the reference is very generic and there
        is unlikely to be semantic issues with reuse or inverse properties,
        one could choose to do something else to satisfy the reference.  This
        would be the case for source_ref/target_ref in plain SROs, for example.

        :param obj: The object to fix up
        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        """

        for parent, key, id_, _ \
                in stix2generator.utils.find_references_assignable(obj):

            # Don't disturb any references already pointing to an existing
            # object.
            if id_ not in by_id:
                stix_type = stix2.utils.get_type_from_id(id_)
                seed_id = self.__make_ref_graph(stix_type, by_id, by_type)
                parent[key] = seed_id

    def __random_get_id(self, by_id, by_type, *required_types):
        """
        Get an ID of a type satisfying the given constraints, admitting
        randomness in whether the ID of an existing object is returned, or
        whether a new object is created.  If a new object is created, its ref
        properties will be filled out according to config settings.

        When setting a ref property, one should be careful: there is potential
        for violating inverse property constraints or other semantic weirdness.
        See __complete_ref_properties().

        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        :param required_types: Type constraints, as a sequence of STIX type
            strings and/or stix2generator.utils.STIXTypeClass enum values.  A
            type constraint must be given, otherwise no satisfying type will
            exist.
        :return: A STIX ID whose type satisfies the given type constraints
        """

        id_ = None
        if random.random() < self.config.probability_reuse:
            id_ = _random_id_of_types(
                by_type, *required_types, stix_version=self.__stix_version
            )

        # It's possible no existing ID satisfies the constraints; create a new
        # object in that case.
        if not id_:
            stix_type = self.__random_generatable_stix_type(*required_types)
            obj = self.__object_generator.generate(stix_type)
            obj_type = obj["type"]
            id_ = obj["id"]

            by_id[id_] = obj
            by_type.setdefault(obj_type, []).append(id_)

            if self.config.complete_ref_properties:
                self.__complete_ref_properties(obj, by_id, by_type)

        return id_

    def __random_ref_list(self, ref_list, by_id, by_type):
        """
        Fill out a reference list from an object, admitting randomness in
        whether IDs of existing objects are used, or whether new objects are
        created.  The graph bookkeeping structures are updated in-place.  Any
        IDs which already refer to graph nodes are left as-is.  Others are
        replaced with IDs of the same type.

        When setting a ref property, one should be careful: there is potential
        for violating inverse property constraints or other semantic weirdness.
        See __complete_ref_properties().

        :param ref_list: A list of references (STIX IDs)
        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        """
        for idx, ref_id in enumerate(ref_list):
            if ref_id not in by_id:

                stix_type = stix2.utils.get_type_from_id(ref_id)

                new_id = self.__random_get_id(by_id, by_type, stix_type)

                ref_list[idx] = new_id

    def __add_sighting(self, by_id, by_type):
        """
        Create a sighting and add it to the graph.  This is not always
        possible, since all new content must connect to existing content to
        avoid a partitioned result, and the "connection points" of a sighting
        (ref properties) have constraints on what they're allowed to refer to.
        Therefore, if there are no objects currently in the graph we can
        connect a sighting to, we can't create a sighting.  This is different
        from a vanilla relationship, for which we always have a generic fallback
        we can use to relate anything ("related-to").  The graph bookkeeping
        structures are updated in-place.

        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        :return: True if a sighting could be created; False if not
        """

        if any(
                stix2generator.utils.is_sdo(type_, self.__stix_version)
                for type_ in by_type
        ):
            sighting = self.__object_generator.generate("sighting")
            observed_data_refs = sighting.get("observed_data_refs")
            where_sighted_refs = sighting.get("where_sighted_refs")

            # We need to connect this sighting to the existing graph, somehow.
            # There are three ways to do that: sighting_of_ref,
            # observed_data_refs, where_sighted_refs.  We know we can connect
            # via sighting_of_ref since we have at least one SDO.  But let's
            # be random about how we connect, as far as that's possible.

            if random.random() < 0.33 \
                    and observed_data_refs \
                    and "observed-data" in by_type:

                # Connect via random element of observed_data_refs
                observed_data_idx = random.randrange(len(observed_data_refs))
                observed_data_id = _random_id_of_types(
                    by_type, "observed-data", stix_version=self.__stix_version
                )
                observed_data_refs[observed_data_idx] = observed_data_id

            elif random.random() < 0.5 \
                    and where_sighted_refs \
                    and ("identity" in by_type or "location" in by_type):

                # Connect via random element of where_sighted_refs
                where_sighted_idx = random.randrange(len(where_sighted_refs))
                where_sighted_id = _random_id_of_types(
                    by_type, "identity", "location",
                    stix_version=self.__stix_version
                )
                where_sighted_refs[where_sighted_idx] = where_sighted_id

            else:

                # Connect via sighting_of_ref
                sighting_of_id = _random_id_of_types(
                    by_type, stix2generator.utils.STIXTypeClass.SDO,
                    stix_version=self.__stix_version
                )
                sighting["sighting_of_ref"] = sighting_of_id

            # Fill out the rest of the references: these randomly may or may
            # not connect to existing graph nodes.
            if sighting["sighting_of_ref"] not in by_id:
                sighting["sighting_of_ref"] = self.__random_get_id(
                    by_id, by_type, stix2generator.utils.STIXTypeClass.SDO
                )

            if observed_data_refs:
                self.__random_ref_list(observed_data_refs, by_id, by_type)

            if where_sighted_refs:
                self.__random_ref_list(where_sighted_refs, by_id, by_type)

            by_id[sighting["id"]] = sighting
            by_type.setdefault("sighting", []).append(sighting["id"])

            if self.config.complete_ref_properties:
                self.__complete_ref_properties(sighting, by_id, by_type)

            success = True

        else:
            # With no SDOs, we can't connect a sighting to any node currently
            # in the graph.  Therefore, we can't create a sighting.
            success = False

        return success

    def __add_sro_reuse(self, by_id, by_type):
        """
        Add a new SRO to the graph, which connects two existing nodes.  The
        bookkeeping structures are updated in-place.

        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        """

        endpoint_id1 = _random_sro_connectable_id(by_type, self.__stix_version)
        endpoint_type1 = stix2.utils.get_type_from_id(endpoint_id1)

        rels = stix2generator.generation.relationships \
            .RELATIONSHIP_OBJECTS_BY_ENDPOINT_TYPE.get(endpoint_type1, [])

        # The idea here is to choose a uniformly random ID over all IDs in the
        # graph connectable to endpoint_id1.  A simpler way would be to
        # first choose a random relationship, then a random ID, but that would
        # not result in a uniform distribution over IDs.
        #
        # reverse_rels_map maps from "other end" type to a list of the
        # relationship objects usable to connect id1 to that other end.
        reverse_rels_map = {}
        for rel in rels:
            endpoint_type2 = rel.target_type \
                if endpoint_type1 == rel.src_type \
                else rel.src_type
            reverse_rels_map.setdefault(endpoint_type2, []).append(rel)

        endpoint_id2 = _random_id_of_types(
            by_type, *reverse_rels_map,
            stix_version=self.__stix_version
        )

        if endpoint_id2:
            endpoint_type2 = stix2.utils.get_type_from_id(endpoint_id2)
            rel = random.choice(reverse_rels_map[endpoint_type2])

            sro = self.__make_sro(
                rel, endpoint_id1, endpoint_id2,
                endpoint_type1, endpoint_type2
            )

        else:
            # We can't find an SRO to use to connect endpoint_id1 to
            # any other graph node.  So connect it to a random node via
            # related-to.
            endpoint_id2 = _random_sro_connectable_id(
                by_type, self.__stix_version
            )
            sro = self.__make_related_to_sro(endpoint_id1, endpoint_id2)

        by_id[sro["id"]] = sro

        if self.config.complete_ref_properties:
            self.__complete_ref_properties(sro, by_id, by_type)

    def __add_sro_new(self, by_id, by_type):
        """
        Add a new SRO to the graph, which connects an existing node to a new
        node.  The bookkeeping structures are updated in-place.

        :param by_id: The current object graph, as a map from ID to object
        :param by_type: Additional bookkeeping: a map from STIX type to a list
            of IDs of STIX objects in the current graph of that type
        """

        endpoint_id1 = _random_sro_connectable_id(by_type, self.__stix_version)
        endpoint_type1 = stix2.utils.get_type_from_id(endpoint_id1)
        rels = stix2generator.generation.relationships \
            .RELATIONSHIP_OBJECTS_BY_ENDPOINT_TYPE.get(endpoint_type1)

        if rels:
            rel = random.choice(rels)

            endpoint_type2 = rel.target_type \
                if endpoint_type1 == rel.src_type \
                else rel.src_type

            endpoint_object2 = self.__object_generator.generate(
                endpoint_type2
            )
            endpoint_id2 = endpoint_object2["id"]

            sro = self.__make_sro(
                rel, endpoint_id1, endpoint_id2,
                endpoint_type1, endpoint_type2
            )

        else:
            # We know of no SROs usable to connect endpoint_id1 to *any* object
            # type.  So just use related-to with a random type.

            endpoint_type2 = self.__random_generatable_stix_type(
                stix2generator.utils.STIXTypeClass.SDO,
                stix2generator.utils.STIXTypeClass.SCO
            )
            endpoint_object2 = self.__object_generator.generate(
                endpoint_type2
            )
            endpoint_id2 = endpoint_object2["id"]

            sro = self.__make_related_to_sro(endpoint_id1, endpoint_id2)

        by_id[endpoint_id2] = endpoint_object2
        by_type.setdefault(endpoint_type2, []).append(endpoint_id2)
        by_id[sro["id"]] = sro

        if self.config.complete_ref_properties:
            self.__complete_ref_properties(sro, by_id, by_type)
            self.__complete_ref_properties(endpoint_object2, by_id, by_type)

    def generate(self, seed_type=None, preexisting_objects=None):
        """
        Generate a STIX graph seeded with an object of the given type.  When
        object reuse is allowed, preexisting_objects provides a way to reuse
        pre-existing objects.  This can be helpful when larger amounts of graph
        content is built up incrementally, so that new regions of content can
        connect with existing regions, instead of each invocation producing its
        own disconnected "island" of content.

        :param seed_type: A STIX type, STIXTypeClass enum, or None.  Generation
            is done by "building out" from a seed object.  This parameter
            determines the type of the seed object.  If None, a STIX SDO is
            chosen at random.
        :param preexisting_objects: The pre-existing STIX content, as either a
            list of objects or mapping from ID to object.
        :return: the generated content, as a mapping from ID to object.  If
            preexisting_objects was given as a map, the same map is returned,
            updated with the new content.  Otherwise, a new map is created and
            returned.  It will contain all of the data from preexisting_objects
            plus the new content.
        """

        if not seed_type:
            seed_type = stix2generator.utils.STIXTypeClass.SDO

        # Might seem kinda silly if seed_type is directly given as a string,
        # but this ensures that the given seed type is actually generatable
        # with our object generator.
        seed_type = self.__random_generatable_stix_type(
            seed_type
        )

        # Pre-populate our data structures, if we were given pre-existing
        # objects.
        #
        # by_id maps from an object ID to an object.  It will hold all objects
        # in the graph.
        if preexisting_objects:
            if isinstance(preexisting_objects, collections.abc.Mapping):
                by_id = preexisting_objects
            else:
                by_id = {
                    obj["id"]: obj
                    for obj in preexisting_objects
                }

        else:
            by_id = {}

        # Wrapping the base preexisting objects map this way allows us to
        # easily distinguish new objects we create from old objects we were
        # given.  At the end, we only want to parse the new objects.
        by_id = collections.ChainMap(by_id).new_child()

        # by_type maps from a STIX type to a list of object IDs in the graph of
        # that type.
        by_type = {}

        for id_, obj in by_id.items():
            obj_type = stix2.utils.get_type_from_id(id_)
            by_type.setdefault(obj_type, []).append(id_)

        seed_obj = self.__object_generator.generate(seed_type)
        seed_id = seed_obj["id"]
        by_id[seed_id] = seed_obj
        by_type.setdefault(seed_type, []).append(seed_id)

        if self.config.complete_ref_properties:
            self.__complete_ref_properties(seed_obj, by_id, by_type)

        # If no SRO-connectable objects available, we can't do anything more!
        if any(
            _is_sro_connectable(type_, self.__stix_version)
            for type_ in by_type
        ):

            num_rels = random.randint(
                self.config.min_relationships, self.config.max_relationships
            )

            for _ in range(num_rels):

                sighting_success = False
                if random.random() < self.config.probability_sighting:
                    sighting_success = self.__add_sighting(by_id, by_type)

                # It's not always possible to add a sighting.  In case we tried
                # and failed, just add a normal SRO instead.
                if not sighting_success:
                    if random.random() < self.config.probability_reuse \
                            and len(by_id) > 1:
                        # Reuse: connect the other end of the SRO to an existing
                        # object.  But if there's only one node, reuse implies
                        # creating a self-loop (reusing that one single node),
                        # which seems weird semantically.  So lets inhibit reuse
                        # in that specific case.

                        self.__add_sro_reuse(by_id, by_type)

                    else:
                        self.__add_sro_new(by_id, by_type)

        # Parse the new objects we created and incorporate them into the base
        # graph.
        new_objects = by_id.maps[0]
        by_id = by_id.maps[1]
        for new_id, new_obj_dict in new_objects.items():
            by_id[new_id] = stix2.parse(
                new_obj_dict, version=self.__stix_version, allow_custom=True
            )

        return by_id



import random
import re

import stix2generator.exceptions
import stix2generator.utils


class ValueConstraintOperators:
    """
    A mixin/base-class for the constraint classes, which contains some
    convenience constants et al for operators.
    """

    # Legal operators
    EQ, NE, GT, GE, LT, LE = OPERATORS = range(6)

    # Reversed operators, aligned with OPERATORS such that the reverse of
    # OPERATORS[i] is REVERSE_OPERATORS[i].  (EQ and NE are symmetric, so they
    # are their own reverses.)
    REVERSE_OPERATORS = EQ, NE, LT, LE, GT, GE

    STRING_OPERATOR_MAP = {
        "=": EQ,
        "!=": NE,
        ">": GT,
        ">=": GE,
        "<": LT,
        "<=": LE
    }

    OPERATOR_STRING_MAP = {
        v: k
        for (k, v) in STRING_OPERATOR_MAP.items()
    }

    @staticmethod
    def to_operator(operator):
        """
        Normalize a value to an operator constant.  If a string is passed,
        convert it to an operator constant.  Otherwise, assume it is already
        an operator constant and just return it unchanged.

        :param operator: An operator string or operator constant
        :return: An operator constant
        """
        if isinstance(operator, str):
            return ValueConstraintOperators.STRING_OPERATOR_MAP[operator]
        else:
            return operator


class ValueConstraint(ValueConstraintOperators):
    """
    Represents a constraint on a property.  E.g. that it be less than some
    value.  Currently this just consists of an operator and a value.  The
    relationship to the property to be constrained is maintained externally,
    so it is not stored in these instances.
    """

    def __init__(self, operator, value):
        self.operator = self.to_operator(operator)
        self.value = value

    def __str__(self):
        return "{} {}".format(
            self.OPERATOR_STRING_MAP[self.operator],
            self.value
        )


class ValueCoconstraint(ValueConstraintOperators):
    """
    Represents a co-constraint between two properties.  This just stores two
    property names and an operator.  When the value of one property becomes
    known, a constraint can be derived for the other property.
    """

    def __init__(self, prop_name_left, operator, prop_name_right):
        self.operator = self.to_operator(operator)
        self.prop_name_left = prop_name_left
        self.prop_name_right = prop_name_right

    def get_constraint(self, prop_name, prop_value):
        """
        Given the value of one property of this co-constraint, derive a
        constraint on the other.

        :param prop_name: The name of one property of this co-constraint
        :param prop_value: The value of prop_name
        :return: A ValueConstraint object representing a constraint on the other
            property.
        :raises ValueError: If the property name given isn't a part of this
            co-constraint.
        """
        if prop_name == self.prop_name_left:
            constraint = ValueConstraint(
                self.REVERSE_OPERATORS[self.operator],
                prop_value
            )

        elif prop_name == self.prop_name_right:
            constraint = ValueConstraint(self.operator, prop_value)

        else:
            # Leave this as a ValueError instead of a ObjectGenerationError
            # subclass.  This error would be the result of an internal bug
            # (i.e. should not happen in a correctly-working library).  Should
            # there be a special InternalError exception class?
            raise ValueError(
                "Property name '{}' not present in coconstraint '{}'".format(
                    prop_name, self
                )
            )

        return constraint

    def involves_property(self, prop_name):
        return prop_name in (self.prop_name_right, self.prop_name_left)

    def get_other_property(self, prop_name):
        if prop_name == self.prop_name_right:
            return self.prop_name_left
        return self.prop_name_right

    def __str__(self):
        return "{} {} {}".format(
            self.prop_name_left,
            self.OPERATOR_STRING_MAP[self.operator],
            self.prop_name_right
        )


class PresenceCoconstraint:

    ONE, ALL, AT_LEAST_ONE = range(3)

    STRING_CONSTRAINT_TYPE_MAP = {
        "one": ONE,
        "all": ALL,
        "at-least-one": AT_LEAST_ONE
    }

    CONSTRAINT_TYPE_STRING_MAP = {
        v: k
        for (k, v) in STRING_CONSTRAINT_TYPE_MAP.items()
    }

    def __init__(self, property_names, constraint_type):
        self.constraint_type = self.to_constraint_type(constraint_type)
        self.property_names = set(property_names)  # dedupes property names

    @staticmethod
    def to_constraint_type(constraint_type):
        if isinstance(constraint_type, str):
            return PresenceCoconstraint.STRING_CONSTRAINT_TYPE_MAP[
                constraint_type
            ]
        else:
            return constraint_type

    def choose_properties(self, probability, minimize_ref_properties):
        """
        Choose some properties from the group, according to its constraint
        type.

        :param probability: This is used only for the at-least-one constraint
            type.  For that type, one is randomly chosen as the required
            property; the rest are optional and included with this probability.
        :param minimize_ref_properties: Modify how property selection is done:
            if True, minimize reference properties in the selection.  Reference
            properties may still be chosen, if there is no other way to
            satisfy the constraint.
        :return: A list of property names.
        """
        assert self.property_names, "Property group must not be empty!"

        non_ref_prop_names = [
            name for name in self.property_names
            if not _is_ref_prop(name)
        ]

        if self.constraint_type == self.ONE:
            if minimize_ref_properties and non_ref_prop_names:
                chosen_props = [random.choice(non_ref_prop_names)]
            else:
                # Or should I have just stored the names as a list...?
                chosen_props = [
                    stix2generator.utils.rand_iterable(self.property_names)
                ]

        elif self.constraint_type == self.AT_LEAST_ONE:

            # Choose one required prop; make the rest optional.  If minimizing
            # reference properties and all are reference properties, we should
            # behave like "one" and choose exactly one.  Choosing more would
            # not be "minimal".
            if minimize_ref_properties and non_ref_prop_names:
                required_prop = random.choice(non_ref_prop_names)
            else:
                required_prop = stix2generator.utils.rand_iterable(
                    self.property_names
                )

            chosen_props = [required_prop]
            candidate_other_props = \
                non_ref_prop_names if minimize_ref_properties \
                else self.property_names

            chosen_props.extend(
                prop for prop in candidate_other_props
                if prop != required_prop and random.random() < probability
            )

        else:
            # ALL
            chosen_props = list(self.property_names)

        return chosen_props

    def can_satisfy_without_refs(self):
        """
        Determine whether it is possible to satisfy this presence co-constraint
        with only non-reference properties.
        """

        if self.constraint_type in (self.ONE, self.AT_LEAST_ONE):
            result = any(not _is_ref_prop(name) for name in self.property_names)

        else:
            # ALL
            result = all(not _is_ref_prop(name) for name in self.property_names)

        return result

    def __str__(self):
        # On python2, this produces the weird set syntax: set(["a","b",...])
        # Should I care about that?
        return "{} {}".format(
            self.CONSTRAINT_TYPE_STRING_MAP[self.constraint_type],
            self.property_names
        )


def _is_ref_prop(name):
    """Determine whether the given name names a "reference" property."""
    return name.endswith("_ref") or name.endswith("_refs")


# This regex just splits a string into an operator and the stuff to its
# left and right.  Trying to avoid a full-fledged parser for this.  It's more
# complicated since I want to avoid an expression like "a!=b" being
# accidentally misinterpreted as "a!" = "b", or "a<=b" misinterpreted as
# "a" < "=b".  So the stuff on the left and right is defined as one or more
# non-operator characters.  The operator starts at the first operator
# character.  An "operator character" is any character appearing in an
# operator.  Operators have symbology which is unlikely to appear in a property
# name, so I hope this is sufficient for now.
_OP_CHARS = "".join(set("".join(ValueConstraintOperators.STRING_OPERATOR_MAP)))
_VALUE_COCONSTRAINT_RE = re.compile(
    "^([^{0}]+)({1})([^{0}]+)$".format(
        re.escape(_OP_CHARS),
        "|".join(
            re.escape(op) for op in ValueConstraintOperators.STRING_OPERATOR_MAP
        )
    ),
    re.S
)


def make_value_coconstraint(value_coconstraint_expr):
    """
    Make a ValueCoconstraint object from a co-constraint expression.  The
    expression syntax is simple: <prop_name> <op> <prop_name>.  E.g. "a < b".

    :param value_coconstraint_expr: The expression, as a string
    :return: The ValueCoconstraint object
    :raises ValueError: if the expression is invalid
    """

    match = _VALUE_COCONSTRAINT_RE.match(value_coconstraint_expr)
    if not match:
        raise stix2generator.exceptions.ValueCoconstraintError(
            value_coconstraint_expr,
            "Invalid expression syntax"
        )

    prop_name_left = match.group(1).strip()
    operator = match.group(2)
    prop_name_right = match.group(3).strip()

    if prop_name_right == prop_name_left:
        raise stix2generator.exceptions.ValueCoconstraintError(
            value_coconstraint_expr,
            "Can't relate a property to itself"
        )

    return ValueCoconstraint(prop_name_left, operator, prop_name_right)



import stix2generator.utils


class GenerationError(Exception):
    """
    Base class for all object and graph generation (via the prototyping
    language) errors.
    """
    pass


class RegistryNotFoundError(GenerationError):
    """
    Represents failure to find an object generator specification registry for
    the given STIX version.
    """
    def __init__(self, stix_version):
        super().__init__(
            "Object specification registry not found for STIX version "
            + stix_version
        )

        self.stix_version = stix_version


class LanguageError(GenerationError):
    """
    Base class for exceptions thrown while processing the STIX prototyping
    language.  Instances may have line/column attributes giving the location
    of the error, if known/applicable.
    """
    def __init__(self, message, parse_tree_meta=None):
        """
        Initialize this error.  If metadata is given, it will be used to
        decorate the message with additional line/column number information.
        The line and column information will also be assigned to same-named
        properties of the exception object.

        :param message: The error message.
        :param parse_tree_meta: Metadata from the parse tree which contains
            line/column number information.  This must be an object with
            "line" and "column" attributes, or None if that information is not
            known or applicable.
        """

        if parse_tree_meta:
            self.line = parse_tree_meta.line
            self.column = parse_tree_meta.column
            message = "{}:{}: {}".format(
                self.line, self.column, message
            )
        else:
            self.line = self.column = None

        super().__init__(message)


class VariableError(LanguageError):
    """
    Instances of this class represent errors regarding variable usage.
    """
    def __init__(self, message, var_name, parse_tree_meta=None):
        """
        Initialize this error.  The name of the variable will be available as
        the var_name property of the exception object.

        :param message: The error message
        :param var_name: The name of the variable.  If this is a Lark Token
            object, it can act as the metadata; parse_tree_meta is not
            necessary in that case.  If parse_tree_meta is given, it takes
            priority.
        :param parse_tree_meta: Metadata from the parse tree, or None.
            See LanguageError for a description of this parameter.
        """
        name_is_token = stix2generator.utils.is_token(var_name)
        if parse_tree_meta is None and name_is_token:
            parse_tree_meta = var_name

        super().__init__(message, parse_tree_meta)

        self.var_name = var_name.value if name_is_token else var_name


class UndeclaredVariableError(VariableError):
    """
    Instances of this class represent use of an undeclared variable.
    """
    def __init__(self, var_name, parse_tree_meta=None):
        """
        Initialize this error.

        :param var_name: The name of the undeclared variable.  See
            VariableError for more description of this parameter.
        :param parse_tree_meta: Metadata from the parse tree, or None.
            See LanguageError for a description of this parameter.
        """
        message = "Undeclared variable: " + var_name
        super().__init__(
            message, var_name, parse_tree_meta
        )


class RedeclaredVariableError(VariableError):
    """
    Instances of this class represent the declaration of a variable more
    than once.
    """
    def __init__(self, var_name, parse_tree_meta=None):
        """
        Initialize this error.

        :param var_name: The name of the redeclared variable.  See
            VariableError for more description of this parameter.
        :param parse_tree_meta: Metadata from the parse tree, or None.
            See LanguageError for a description of this parameter.
        """
        message = "Redeclared variable: " + var_name
        super().__init__(
            message, var_name, parse_tree_meta
        )


class CircularVariableDependenciesError(LanguageError):
    """
    Instances of this class represent circular dependencies among variables.
    """
    def __init__(self, path):
        """
        Initialize this error.  The path will be available as the same-named
        property of the exception object.

        :param path: A list of strings which gives the circular variable
            dependency path which was found.
        """
        super().__init__(
            "Circular dependency detected in variable declarations: "
            + " > ".join(path)
        )

        self.path = path


class ObjectGenerationError(GenerationError):
    """
    Base class for object generation errors.  Also used directly as a raised
    exception type, when there isn't a more specific object generation error
    subclass that's applicable.  This is typically used for various types of
    problems with specifications.

    This class is written (and used) in a different way, to make it easier to
    have good error messages.  I think there is a relatively common problem in
    exception handling, that at the point where an exception object is created,
    not all contextual information is known.  Python's default behavior
    requires that a complete message be generated at the point the exception
    object is constructed.  When you don't have all the contextual information
    at that point, it's just impossible to generate a good message.  That makes
    it tempting to pass extra context junk into functions/methods which is only
    used to generate decent error messages.  In simple cases, that may be fine,
    but in excess, it can clutter the API.

    An approach to improving the situation is to allow an exception to have
    some extra info added as it propagates through stack frames where that
    info is known.  I first came across this idea in Boost years ago, and I
    decided to try a similar design here.  My idea is that exception handlers
    can specifically add specification type and name info, if known, as the
    exception propagates.  Then I override how the message is formatted to
    incorporate this additional information.
    """

    def __init__(self, message, spec_type=None, spec_name_stack=None):
        """
        Initialize the exception.

        :param message: The base exception message (no need to include spec
            type or name information in the message; you can pass those in
            separately and the message will be augmented to include that info
            in a standard way).
        :param spec_type: A spec type if known/applicable, or None if not.
            E.g. a JSON type or semantic name.
        :param spec_name_stack: A spec name or list of names if
            known/applicable, or None if not.  This describes a chain of
            references.  For convenience, a plain name will be treated as a
            length-1 list.
        """

        super().__init__(message)

        self.__message = message

        # attrs for exception handlers to set
        self.spec_type = spec_type
        if isinstance(spec_name_stack, str):
            self.spec_name_stack = [spec_name_stack]
        else:
            self.spec_name_stack = spec_name_stack or []

    def __str__(self):
        """
        Overrides the base class's exception message generation.  This uses
        the spec name/type attrs to improve the message, if they were set.
        """
        msg_parts = []
        if self.spec_name_stack:
            stack_desc = " > ".join(
                "'{}'".format(name) for name in self.spec_name_stack
            )

            msg_parts.append(
                "In specification {}: ".format(stack_desc)
            )

        if self.spec_type:
            msg_parts.append(
                "Error generating {}: ".format(self.spec_type)
            )

        msg_parts.append(self.__message)

        return "".join(msg_parts)


class UnrecognizedJSONTypeError(ObjectGenerationError):
    """
    Represents an unrecognized spec type, e.g. in a spec's "type" property.
    """
    def __init__(self, spec_type, spec_name_stack=None):

        super().__init__(
            "unrecognized JSON type", spec_type, spec_name_stack
        )


class CyclicSpecificationReferenceError(ObjectGenerationError):
    """
    Instances represent a reference cycle composed of specification names.
    """
    def __init__(self, spec_name_cycle, spec_type=None, spec_name_stack=None):

        message = "Specification reference cycle detected: " + \
            " > ".join(spec_name_cycle)

        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.spec_name_cycle = spec_name_cycle


class TypeMismatchError(ObjectGenerationError):
    """
    Instances represent a spec type mismatch: a spec for a particular JSON
    type was required, but a spec for a different type was found instead.
    """
    def __init__(self, expected_type, actual_type, spec_type=None,
                 spec_name_stack=None):

        message = "Type mismatch: expected '{}' but got '{}'".format(
            expected_type, actual_type
        )

        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.expected_type = expected_type
        self.actual_type = actual_type


class SemanticValueTypeMismatchError(ObjectGenerationError):
    """
    Instances represent a mismatch between the type of value produced by a
    semantic implementation, and the declared type of a specification.
    """

    def __init__(self, semantic_name, actual_type, actual_value, spec_type,
                 spec_name_stack=None):

        message = "Semantic '{}' produced a value of the wrong type: " \
                  "expected {}, got {}: {}".format(
                      semantic_name, spec_type, actual_type, actual_value
                  )

        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.semantic_name = semantic_name
        self.actual_type = actual_type
        self.actual_value = actual_value


class SpecificationNotFoundError(ObjectGenerationError):
    """
    A specification referred to by name was not found in the registry.
    """
    def __init__(self, spec_name, spec_type=None, spec_name_stack=None):

        message = "Spec not found: '{}'".format(spec_name)

        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.spec_name = spec_name


class UndefinedPropertyError(ObjectGenerationError):
    """
    Instances represent a reference outside a co-constraint to a property which
    was not defined in an object specification.  (Inside a co-constraint, a
    different co-constraint error type is used, to indicate a bad
    co-constraint.)
    """
    def __init__(self, prop_names, spec_type=None, spec_name_stack=None):

        # Support a single string as well, converting to a list
        if isinstance(prop_names, str):
            prop_names = [prop_names]

        message = "Reference to undefined property(s): " + \
                  ", ".join(prop_names)
        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.prop_names = prop_names


class ValueCoconstraintError(ObjectGenerationError):
    """
    Instances represent an invalid value co-constraint.
    """
    def __init__(self, coconstraint, message, spec_type=None,
                 spec_name_stack=None):

        message = "Invalid value co-constraint '{}': {}".format(
            coconstraint,
            message
        )

        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.coconstraint = coconstraint


class PresenceCoconstraintError(ObjectGenerationError):
    """
    Instances represent an invalid presence co-constraint.
    """
    pass


class InvalidPropertyGroupError(PresenceCoconstraintError):
    """
    Instances represent an invalid property group, within a presence
    co-constraint definition.
    """
    def __init__(self, group_name, message, spec_type=None,
                 spec_name_stack=None):

        message = 'Invalid property group "{}": {}'.format(group_name, message)
        super().__init__(
            message, spec_type, spec_name_stack
        )

        self.group_name = group_name


class GeneratableSTIXTypeNotFoundError(GenerationError):
    """
    Instances of this class represent inability to find a generator spec
    registered with an object generator, which satisfies particular
    STIX-related criteria.  For example, we might require any SDO or SCO type.
    """
    def __init__(self, constraints, stix_version):
        """
        Initialize this exception instance.

        :param constraints: The constraints which failed to match any
            specification.  Must be a list of constraint values.  See
            stix2generator.utils.random_generatable_stix_type() or
            .is_stix_type() for more information.
        :param stix_version: The STIX version which the constraints were
            checked relative to
        """

        message = "Could not find an object generator specification for a " \
                  "STIX {} type satisfying the constraints: {}".format(
                      stix_version,
                      constraints
                  )
        super().__init__(message)

        self.constraints = constraints
        self.stix_version = stix_version


class PatternGenerationError(GenerationError):
    """
    Base class for errors related to random STIX pattern generation.
    """
    pass


class UnhandledPropertyValueType(PatternGenerationError):
    """
    Instances represent an unhandled property value type from a stix2 object.
    The property value is used to create a constant in a STIX pattern, and we
    need to know what comparison expression operators are legal to use with it.
    """
    def __init__(self, value):
        message = "Can't create comparison expression: don't know what " \
                  "operators apply to type {}: {}".format(
                      type(value).__name__, str(value)
                  )
        super().__init__(message)

        self.value = value


class UnsupportedObjectStructureError(PatternGenerationError):
    """
    The AST for object paths can't represent arbitrary path structure.  E.g. it
    can't represent a list of lists.  Instances of this class represent a
    random path taken through a STIX object, where the path contains
    unsupported structure.
    """
    def __init__(self, object_type, path_elements):
        message = "The path through a '{}' contains unsupported structure:" \
                  " {}".format(
                      object_type, repr(path_elements)
                  )
        super().__init__(message)

        self.object_type = object_type
        self.path_elements = path_elements


class UnrecognizedSTIXTypeError(PatternGenerationError):
    """
    Instances represent a STIX SCO type given as a comparison expression type
    constraint, which was unrecognized.  A generator spec could not be found
    which generates objects of that type.
    """
    def __init__(self, stix_type):
        message = "Unrecognized STIX type for comparison expression type" \
                  " constraint: {}".format(stix_type)
        super().__init__(message)

        self.stix_type = stix_type


class InvalidRefPropertyValueError(PatternGenerationError):
    """
    Instances represent an invalid value for a reference property.  The value
    must be an ID from which we can extract an object type, which is used to
    continue an object path through the reference.
    """
    def __init__(self, value):
        message = "Invalid reference property value, must be an" \
                  " object ID: {}".format(value)

        super().__init__(message)
        self.value = value


class AutoRegistrationError(Exception):
    """
    Base class for errors with automatic creation and registration of
    Python classes as stix2 custom objects, which are derived from object
    generator specifications.
    """
    def __init__(self, message, spec_name=None):
        super().__init__(message)

        self.__message = message
        self.spec_name = spec_name

    def __str__(self):
        msg_parts = []
        if self.spec_name:
            msg_parts.append(
                "In specification {}: ".format(self.spec_name)
            )

        msg_parts.append(self.__message)

        return "".join(msg_parts)


class AutoRegistrationInferenceError(AutoRegistrationError):
    """
    For auto-registration to work, STIX object information must be inferred
    from an object generator specification.  Instances of this class represent
    situations where the spec is such that inference can't be performed (even
    if the spec itself is valid).
    """
    pass


class IllegalSTIXObjectSpecType(AutoRegistrationInferenceError):
    """
    Specifications for STIX objects must be of type "object".
    """

    def __init__(self, spec_type, spec_name=None):
        message = 'STIX object specs must have type "object"; got "{}"'.format(
            spec_type
        )
        super().__init__(message, spec_name)

        self.spec_type = spec_type


class IllegalSTIXObjectPropertyType(AutoRegistrationInferenceError):
    """
    An object generator property spec was for a JSON type which is unsupported
    for use as a stix2 property type.
    """
    def __init__(self, prop_type, spec_name=None):
        message = 'Don\'t know how to make a stix2 property object for ' \
                  'spec type "{}"'.format(prop_type)
        super().__init__(message, spec_name)

        self.prop_type = prop_type


class EmptyListError(AutoRegistrationInferenceError):
    def __init__(self):
        message = "Can't infer a list type from an empty list.  The list's " \
                  "element type must be inferrable, so it needs at least one " \
                  "value."
        super().__init__(message)


class HeterogenousListError(AutoRegistrationInferenceError):
    def __init__(self, list_):
        message = "Can't infer a list element type from a heterogenous " \
                  "list: " + str(list_)
        super().__init__(message)

        self.list_ = list_



import collections


# Summary of defined SROs.  This basically shows how STIX types are allowed to
# relate to each other, and via which relationship types.
#
# Hopefully this is an easy structure to maintain: maps relationship source
# STIX type to a map from relationship type to a list of target types.  It
# eliminates repetition in source and relationship types.
_RELATIONSHIPS = {
    "attack-pattern": {
        "delivers": ["malware"],
        "targets": ["identity", "location", "vulnerability"],
        "uses": ["malware", "tool"]
    },
    "campaign": {
        "attributed-to": ["intrusion-set", "threat-actor"],
        "compromises": ["infrastructure"],
        "originates-from": ["location"],
        "targets": ["identity", "location", "vulnerability"],
        "uses": ["attack-pattern", "infrastructure", "malware", "tool"]
    },
    "course-of-action": {
        "investigates": ["indicator"],
        "mitigates": [
            "attack-pattern", "indicator", "malware", "tool", "vulnerability"
        ]
    },
    "identity": {
        "located-at": ["location"]
    },
    "indicator": {
        "indicates": [
            "attack-pattern", "campaign", "infrastructure", "intrusion-set",
            "malware", "threat-actor", "tool"
        ],
        "based-on": ["observed-data"]
    },
    "infrastructure": {
        "communicates-with": [
            "infrastructure", "ipv4-addr", "ipv6-addr", "domain-name", "url"
        ],
        "consists-of": [
            "infrastructure", "observed-data",
            # all generatable SCO types
            "artifact", "autonomous-system", "directory", "domain-name",
            "email-addr", "email-message", "file", "ipv4-addr", "ipv6-addr",
            "mac-addr", "mutex", "network-traffic", "process", "software",
            "url", "user-account", "windows-registry-key", "x509-certificate"
        ],
        "controls": ["infrastructure", "malware"],
        "delivers": ["malware"],
        "has": ["vulnerability"],
        "hosts": ["tool", "malware"],
        "located-at": ["location"],
        "uses": ["infrastructure"],
    },
    "intrusion-set": {
        "attributed-to": ["threat-actor"],
        "compromises": ["infrastructure"],
        "hosts": ["infrastructure"],
        "owns": ["infrastructure"],
        "originates-from": ["location"],
        "targets": ["identity", "location", "vulnerability"],
        "uses": ["attack-pattern", "infrastructure", "malware", "tool"]
    },
    "malware": {
        "authored-by": ["threat-actor", "intrusion-set"],
        "beacons-to": ["infrastructure"],
        "exfiltrate-to": ["infrastructure"],
        "communicates-with": ["ipv4-addr", "ipv6-addr", "domain-name", "url"],
        "controls": ["malware"],
        "downloads": ["malware", "tool", "file"],
        "drops": ["malware", "tool", "file"],
        "exploits": ["vulnerability"],
        "originates-from": ["location"],
        "targets": ["identity", "infrastructure", "location", "vulnerability"],
        "uses": ["attack-pattern", "infrastructure", "malware", "tool"],
        "variant-of": ["malware"],
    },
    "malware-analysis": {
        "characterizes": ["malware"],
        "analysis-of": ["malware"],
        "static-analysis-of": ["malware"],
        "dynamic-analysis-of": ["malware"]
    },
    "threat-actor": {
        "attributed-to": ["identity"],
        "compromises": ["infrastructure"],
        "hosts": ["infrastructure"],
        "owns": ["infrastructure"],
        "impersonates": ["identity"],
        "located-at": ["location"],
        "targets": ["identity", "location", "vulnerability"],
        "uses": ["attack-pattern", "infrastructure", "malware", "tool"]
    },
    "tool": {
        "delivers": ["malware"],
        "drops": ["malware"],
        "has": ["vulnerability"],
        "targets": ["identity", "infrastructure", "location", "vulnerability"]
    }
}


# Programmatically add some common relationships, to reduce verbosity in
# the _RELATIONSHIPS map
for src_type, rel_info in _RELATIONSHIPS.items():
    rel_info["derived-from"] = [src_type]
    rel_info["duplicate-of"] = [src_type]


# Omit "related-to".  Currently special-casing that in the STIX generator.


Relationship = collections.namedtuple(
    "Relationship", "src_type, rel_type, target_type"
)


# A structure to make it easy to choose a graph edge based on the type of node
# it needs to connect to at one end.
RELATIONSHIP_OBJECTS_BY_ENDPOINT_TYPE = {}


for src_type, rel_info in _RELATIONSHIPS.items():
    for rel_type, target_types in rel_info.items():
        for target_type in target_types:

            rel_obj = Relationship(src_type, rel_type, target_type)

            RELATIONSHIP_OBJECTS_BY_ENDPOINT_TYPE.setdefault(
                src_type, []
            ).append(rel_obj)

            RELATIONSHIP_OBJECTS_BY_ENDPOINT_TYPE.setdefault(
                target_type, []
            ).append(rel_obj)
